{-# OPTIONS -fglasgow-exts #-}
module Heap where

--------------------------------------------------------------------------
-- imports

import Property
import Test.QuickCheck hiding (forAll)

import Data.List
  ( sort
  , (\\)
  )

import Control.Monad
  ( liftM2
  )

--------------------------------------------------------------------------
-- skew heaps

data Heap a
  = Node a (Heap a) (Heap a)
  | Empty
 deriving ( Eq, Ord, Show )
  
empty :: Heap a
empty = Empty

isEmpty :: Heap a -> Bool
isEmpty Empty = True
isEmpty _     = False

unit :: a -> Heap a
unit x = Node x empty empty

size :: Heap a -> Int
size Empty          = 0
size (Node _ h1 h2) = 1 + size h1 + size h2

insert :: Ord a => a -> Heap a -> Heap a
insert x h = unit x `merge` h

removeMin :: Ord a => Heap a -> Maybe (a, Heap a)
removeMin Empty          = Nothing
removeMin (Node x h1 h2) = Just (x, h1 `merge` h2)

merge :: Ord a => Heap a -> Heap a -> Heap a
h1    `merge` Empty = h1
Empty `merge` h2    = h2
h1@(Node x h11 h12) `merge` h2@(Node y h21 h22)
  | x <= y    = Node x (h12 `merge` h2) h11
  | otherwise = Node y (h22 `merge` h1) h21
        
fromList :: Ord a => [a] -> Heap a
fromList xs = merging [ unit x | x <- xs ]
 where
  merging []  = empty
  merging [h] = h
  merging hs  = merging (sweep hs) 

  sweep []         = []
  sweep [h]        = [h]
  sweep (h1:h2:hs) = (h1 `merge` h2) : sweep hs

toList :: Heap a -> [a]
toList h = toList' [h]
 where
  toList' []                  = []
  toList' (Empty        : hs) = toList' hs
  toList' (Node x h1 h2 : hs) = x : toList' (h1:h2:hs)

toSortedList :: Ord a => Heap a -> [a]
toSortedList Empty          = []
toSortedList (Node x h1 h2) = x : toList (h1 `merge` h2)

-- The above contains a bug. The correct last line of that definition
-- is:
-- toSortedList (Node x h1 h2) = x : toSortedList (h1 `merge` h2)

--------------------------------------------------------------------------
-- properties

-- (<=?) :: Ord a => a -> Heap a -> Prop
-- x <=? Empty      = true
-- x <=? Node y _ _ = holds (x <= y)
-- 
-- (==?) :: Ord a => Heap a -> [a] -> Prop
-- h ==? xs = invariant h /\ sort (toList h) === sort xs
-- 
-- invariant :: Ord a => Heap a -> Prop
-- invariant Empty          = true
-- invariant (Node x h1 h2) = x <=? h1 /\ x <=? h2 /\ invariant h1 /\ invariant h2

(<=?) :: Ord a => a -> Heap a -> Bool
x <=? Empty      = True
x <=? Node y _ _ = x <= y

(==?) :: Ord a => Heap a -> [a] -> Bool
h ==? xs = invariant h && sort (toList h) == sort xs

invariant :: Ord a => Heap a -> Bool
invariant Empty          = True
invariant (Node x h1 h2) = x <=? h1 && x <=? h2 && invariant h1 && invariant h2

prop_Empty =
  holds (empty ==? ([] :: [Int]))

prop_Unit =
  forAll $ \(x :: Int) -> holds (unit x ==? [x])

prop_Size =
  forAll $ \(h :: Heap Int) -> size h === length (toList h)

prop_Insert =
  forAll $ \x -> forAll $ \(h :: Heap Int) -> holds (insert x h ==? (x : toList h))

prop_removeMin_aux h Nothing = holds (h ==? [])
prop_removeMin_aux h (Just (x,h')) =
    x === minimum (toList h) /\ holds (h' ==? (toList h \\ [x]))

prop_RemoveMin =
  forAll $ \(h :: Heap Int) -> prop_removeMin_aux h (removeMin h)

prop_Merge =
  forAll $ \h1 -> forAll $ \(h2 :: Heap Int) ->
  holds $ (h1 `merge` h2) ==? (toList h1 ++ toList h2)

prop_FromList =
  forAll $ \(xs :: [Int]) ->
  holds $ fromList xs ==? xs

prop_ToSortedList =
  forAll $ \(h :: Heap Int) ->
  let xs = toSortedList h in
  holds (h ==? xs) /\ xs === sort xs
  
--------------------------------------------------------------------------
-- QuickCheck specific

instance (Ord a, Arbitrary a) => Arbitrary (Heap a) where
  arbitrary = sized (arbHeap Nothing)
   where
    arbHeap mx n =
      frequency $
        [ (1, return Empty) ] ++
        [ (7, do my <- arbitrary `suchThatMaybe` ((>= mx) . Just)
                 case my of
                   Nothing -> return Empty
                   Just y  -> liftM2 (Node y) arbHeap2 arbHeap2
                    where arbHeap2 = arbHeap (Just y) (n `div` 2))
        | n > 0
        ]
  coarbitrary = undefined

suchThatMaybe :: Gen a -> (a -> Bool) -> Gen (Maybe a)
gen `suchThatMaybe` p = sized tries
 where
  tries 0 = do return Nothing
  tries n = do a <- gen
               if p a then return (Just a)
                      else tries (n-1)

{-
  shrink Empty          = []
  shrink (Node x h1 h2) = [ h1, h2 ]
                       ++ [ Node x h1' h2 | h1' <- shrink h1, x <=? h1' ]
                       ++ [ Node x h1 h2' | h2' <- shrink h2, x <=? h2' ]
                       ++ [ Node x' h1 h2 | x' <- shrink x, x' <=? h1, x' <=? h2 ]
-}

-- quickCheckAll =
--   do quickCheck prop_Empty
--      quickCheck prop_Unit
--      quickCheck prop_Size
--      quickCheck prop_Insert
--      quickCheck prop_RemoveMin
--      quickCheck prop_Merge
--      quickCheck prop_FromList
--      quickCheck prop_ToSortedList

--------------------------------------------------------------------------
-- the end.
