{-| Abstracting a value @w@ from another value @v@ is the process of
    finding a function @f@ such that @f w = v@.
-}

module TypeChecker.Abstraction where

import Control.Monad.State

import Syntax.Internal

import TypeChecker.UnfoldConstants
import TypeChecker.Universes
import TypeChecker.Values

import Utils.Monads
--import Utils.Debug

-- * Abstraction

-- | We should probably order the results cleverly.
abstract :: Value -> Value -> [Value -> Value]
abstract w v =
	evalState (abstr w v) 0
    where
	mkLam f = VLam (Cl "hole" f) undefined
	abstr w v =
	    ifM (valueType w `subtype'` valueType v &&& v `equal` w)
		(return [ id, const v ])
	    $ case unfoldConstants v of
		VApp v v'	->
		    do	fs <- abstr w v
			gs <- abstr w v'
			return [ \v -> f v `vApp` g v | f <- fs, g <- gs ]
		VLam cl t	->
		    do	let ~(VPi t' _ _) = unfoldConstants t
			u@(VId x _) <- freshValue t'
			fs  <- abstr w (cl `clApp` u)
			return	[ \v -> VLam (cl{ clApp = \w ->
						    substitute x w (f (VTag v))
						}) t
				| f <- fs
				]
		VPi v' cl t	->
		    do	u@(VId x _) <- freshValue w
			fs	    <- abstr w v'
			gs	    <- abstr w (cl `clApp` u)
			return	[ \v -> VPi (f v)
					    (cl{ clApp = \w ->
						    substitute x w (g (VTag v))
						}) t
				| f <- fs
				, g <- gs
				]
		v -> return [ const v ]

-- * Equality

freshValue t =
    do	n <- get
	put (n + 1)
	return $ VId ("@" ++ show n) t

infixr 2 &&&
(&&&) :: Monad m => m Bool -> m Bool -> m Bool
m1 &&& m2 = ifM m1 m2 (return False)

(=<=) :: Value -> Value -> Bool
v =<= w	= evalState (v `subtype'` w) 0

subtype' v w =
    case (unfoldConstants v, unfoldConstants w) of
	(VSort s1, VSort s2)		    -> return (s1 <= s2)
	_   | valueType v =<= valueType w   -> v `subtype` w
	    | otherwise			    -> return False

v `subtype` w = unfoldConstants v =<=* unfoldConstants w

VSort s1 =<=* VSort s2 = return (s1 <= s2)

VDef x _ =<=* VDef y _
    | x == y	= return True

-- ? Alternative: v === w, theoretical implications ?
VPi v f _ =<=* VPi w g _ =
    w `subtype` v &&&
	do  x <- freshValue w
	    clApp f x `subtype` clApp g x

v =<=* w = v `equal` w



(===) :: Value -> Value -> Bool
v === w = evalState (equal v w) 0

equal :: Value -> Value -> State Int Bool

v1 `equal` v2 = unfoldConstants v1 ==* unfoldConstants v2

VDef x _    ==* VDef y _    = return (x == y)
VId x _	    ==* VId y _	    = return (x `eqVar` y)
    where
	"_" `eqVar` _	= False
	_ `eqVar` "_"	= False
	x `eqVar` y	= x == y

VLam f t ==* VLam g _ =
    do	let ~(VPi w _ _) = unfoldConstants t
	x <- freshValue w
	clApp f x `equal` clApp g x

VLam f t ==* u = 
    do	let ~(VPi w _ _) = unfoldConstants t
	x <- freshValue w
	clApp f x `equal` vApp u x

u ==* VLam f t =
    do	let ~(VPi w _ _) = unfoldConstants t
	x <- freshValue w
	VApp u x `equal` clApp f x

VPi v f _ ==* VPi w g _ =
    v `equal` w &&& do	x <- freshValue v
			clApp f x `equal` clApp g x

VSort s1 ==* VSort s2 =
    return (s1 == s2)

VApp u1 v1 ==* VApp u2 v2 =
    u1 `equal` u2 &&& v1 `equal` v2

VTop _ ==* VTop _ = return True

u ==* v = return False

-- * Substitution

{-| Danger:
	Substitution under lambdas should not substitute in the
	function argument. Passing around a substitution should
	do the trick. I can't construct a failing case at the
	moment, but it should be possible.
-}
substitute :: Name -> Value -> Value -> Value
substitute x u v =
    case v of
	VId y _	| x == y    -> u
	VApp v w	    -> sub v `vApp` sub w
	VLam f t	    -> VLam (subCl f) t
	VPi w f t	    -> VPi (sub w) (subCl f) t
	VTag v		    -> v
	VId _ _		    -> v
	VDef _ _	    -> v
	VSort _		    -> v
	VTop _		    -> v
--	_		    -> error $ "Internal error in substitute: " ++ show v
    where
	sub	= substitute x u
	subCl f = f { clApp = \v -> sub (f `clApp` VTag v) }

